import numpy as np
import numpy.polynomial.chebyshev as cheb
import scipy.sparse as sp
from scipy import interp
from scipy.interpolate import interp1d

# supress warnings
import warnings
warnings.filterwarnings('ignore')

# hard coded algo params
m_lo = 0.5 # k_lo multiplier
m_hi = 2.0 # k_hi multiplier
cmin = 1e-5 # minimum allowed consumption

# sigmoid function
sigmoid = lambda x: 1/(1+np.exp(-x))

# specialized interpolator
def interpf(grid, val, kind='quadratic'):
    return interp1d(grid, val, kind=kind, bounds_error=False, fill_value='extrapolate')

##
## CRRA utility function
##

# CRRA utility function
def util(c, sig):
    c1 = np.maximum(cmin, c)
    if sig == 1:
        return np.log(c1)
    else:
        return (c1**(1-sig)-1)/(1-sig)

# inverse of derivative of CRRA utility function
def dutil_inv(dv, sig):
    dv1 = np.maximum(cmin, dv)
    return dv1**(-1/sig)

##
## vector tools
##

# make sure it's a vector
def ensure_vector(x, N):
    if np.isscalar(x):
        return x*np.ones(N)
    else:
        return np.array(x)

# like choose but d omits axis of choice
def address0(d, f):
    sd0 = f.shape[0]
    sd1 = np.prod(f.shape[1:])
    ic = sd1*d.reshape((-1,)) + np.arange(sd1)
    return f.flat[ic].reshape(d.shape)

# address along a given axis
def address(d, f, axis=0):
    # swap to 0th axis
    if axis != 0:
        f = f.swapaxes(0, axis)

    # broadcast trailing axes
    if f.shape[1:] != d.shape:
        f = np.broadcast_to(f, (f.shape[0],)+d.shape)

    # apply with 0th version
    g = address0(d, f)

    return g

# golden section search over vector
from scipy.constants import golden
def goldsec_vec(fun, xmin, xmax, tol=1e-12):
    val0, val1 = fun(xmin), fun(xmax) # testing sizes
    N = len(val0)
    assert(len(val1) == N)

    a = ensure_vector(xmin, N)
    b = ensure_vector(xmax, N)

    c = b - (b-a)/golden
    d = a + (b-a)/golden

    while np.max(np.abs(c-d)) > tol:
        sel = fun(c) < fun(d)
        b[sel] = d[sel]
        a[~sel] = c[~sel]

        c = b - (b-a)/golden
        d = a + (b-a)/golden

    return (b+a)/2

##
## Discrete Time
##

# using grid search
def disc_grid(beta=0.95, alpha=0.3, delta=0.1, sigma=2.0,
              N=512, K=5000, drag=0, tol=1e-12, output=False):
    # steady state
    k_ss = ((alpha*beta)/(1-beta*(1-delta)))**(1/(1-alpha))

    # grid it up
    k_lo, k_hi = m_lo*k_ss, m_hi*k_ss
    k_grid = np.linspace(k_lo, k_hi, N)

    # extras stuff
    y = k_grid**alpha
    inc = y + (1-delta)*k_grid

    # (row,column) = (t,t+1)
    c = inc[:, None] - k_grid[None, :]
    u = util(c, sig=sigma)

    # initial guess okay
    val = np.zeros(N)

    # iterate until convergence
    for i in range(K):
        val_mat = u + beta*val[None, :]
        kpi_opt = np.argmax(val_mat, axis=1)
        val_new = address(kpi_opt, val_mat, axis=1)

        # check for convergence
        err = np.max(np.abs(val_new-val))
        if err < tol:
            if output:
                print(f'{i}: {err:8g}')
                print('converged')
            break

        # progress update
        if output and i % 500 == 0:
            print(f'{i}: {err:8g}')

        # update value
        val = drag*val + (1-drag)*val_new

    # finish up
    kp_opt = k_grid[kpi_opt]
    vp_opt = val[kpi_opt]
    c_opt = inc - kp_opt
    u_opt = util(c_opt, sig=sigma)

    # return vectors
    return {
        'k': k_grid,
        'v': val,
        'kp': kp_opt,
        'vp': vp_opt,
        'c': c_opt,
        'u': u_opt
    }

# using interpolation
def disc_interp(beta=0.95, alpha=0.3, delta=0.1, sigma=2.0,
                N=512, K=5000, drag=0, tol=1e-12, output=False):
    # steady state
    k_ss = ((alpha*beta)/(1-beta*(1-delta)))**(1/(1-alpha))

    # grid it up
    k_lo, k_hi = m_lo*k_ss, m_hi*k_ss
    k_grid = np.linspace(k_lo, k_hi, N)

    # precompute
    y = k_grid**alpha
    inc = y + (1-delta)*k_grid

    # initial guess okay
    val = np.zeros(N)

    # iterate until convergence
    for i in range(K):
        valf = interpf(k_grid, val, kind='quadratic')
        vobj = lambda x: util(inc-x, sig=sigma) + beta*valf(x)
        kp_opt = goldsec_vec(lambda x: -vobj(x), 0, inc)
        val_new = vobj(kp_opt)

        # check for convergence
        err = np.max(np.abs(val_new-val))
        if err < tol:
            if output:
                print(f'{i}: {err:8g}')
                print('converged')
            break

        # progress update
        if output and i % 500 == 0:
            print(f'{i}: {err:8g}')

        # update value
        val = drag*val + (1-drag)*val_new

    # finish up
    vp_opt = valf(kp_opt)
    c_opt = inc - kp_opt
    u_opt = util(c_opt, sig=sigma)

    # return vectors
    return {
        'k': k_grid,
        'v': val,
        'kp': kp_opt,
        'vp': vp_opt,
        'c': c_opt,
        'u': u_opt
    }

# check errors on a higher interpolated grid
def disc_errors(k0_grid, val, kp, N1, kind='linear',
                beta=0.95, alpha=0.3, delta=0.1, sigma=2.0):
    # steady state
    k_ss = ((alpha*beta)/(1-beta*(1-delta)))**(1/(1-alpha))
    k_lo, k_hi = m_lo*k_ss, m_hi*k_ss

    # grid it up
    k1_grid = np.linspace(k_lo, k_hi, N1)

    # interpolated values
    val_int = interp(k1_grid, k0_grid, val)
    kp_int = interp(k1_grid, k0_grid, kp)
    vp_int = interp(kp_int, k0_grid, val)

    # equation error
    y1 = k1_grid**alpha
    inc1 = y1 + (1-delta)*k1_grid
    val1 = util(inc1-kp_int, sig=sigma) + beta*vp_int
    verr = val_int - val1

    return k1_grid, verr

##
## Continuous Time
##

# potentially limited update
def update(vold, vnew, drag, upd):
    vmin, vmax = vold - upd, vold + upd
    vupd = drag*vold + (1-drag)*vnew
    vcut = np.maximum(vmin, np.minimum(vmax, vupd))
    return vcut

# proper winding scheme
def upwind(val, diff, direc):
    dval0 = (val[1:]-val[:-1])/diff
    dval_hi = np.r_[dval0, dval0[-1]]
    dval_lo = np.r_[dval0[0], dval0]
    sign = sigmoid(direc)
    dval = sign*dval_hi + (1-sign)*dval_lo
    return dval

# using rediscretization
def cont_redisc(rho=0.05, alpha=0.3, delta=0.1, sigma=2.0,
                N=128, K=500000, delt=0.01, tol=1e-12, skip=200,
                steep=200, drag=0.5, upd=0.1, output=False):
    # steady state
    k_ss = (alpha/(rho+delta))**(1/(1-alpha))
    s_ss = alpha*(delta/(rho+delta))

    # grid it up
    k_lo, k_hi = m_lo*k_ss, m_hi*k_ss
    k_grid = np.linspace(k_lo, k_hi, N)
    k_diff = k_grid[1:] - k_grid[:-1]

    # extras stuff
    disc = np.exp(-rho*delt)
    y = k_grid**alpha

    # initial guess
    inv = (s_ss*y)*np.ones(N) # k_dot = 0
    val = util(y, sig=sigma)*np.ones(N) # inv = 0

    # iterate until convergence
    for i in range(K):
        # find derivative
        k_dot = inv - delta*k_grid
        dval = upwind(val, k_diff, steep*k_dot)

        # update value and investment (independently)
        val_new = delt*util(y-inv, sig=sigma) + disc*(val+delt*k_dot*dval)
        inv_new = y - dutil_inv(disc*dval, sig=sigma)

        # check for convergence
        val_err = np.max(np.abs(val_new-val))
        inv_err = np.max(np.abs(inv_new-inv))
        err = np.max([val_err, inv_err])
        if err < tol:
            if output:
                print(f'{i}: {err:8g}')
                print('converged')
            break

        # progress update
        if output and i % 10000 == 0:
            print(f'{i}: {err:8g}')

        # update value (only update inv periodically)
        val = update(val, val_new, drag, upd)
        if i % skip == 0:
            inv = update(inv, inv_new, drag, upd)

    # finish up
    c = y - inv
    u = util(c, sig=sigma)
    kdot = inv - delta*k_grid

    # return vectors
    return {
        'k': k_grid,
        'v': val,
        'i': inv,
        'dk': kdot,
        'dv': dval,
        'c': c,
        'u': u
    }

def cont_sparse(rho=0.05, alpha=0.3, delta=0.1, sigma=2.0,
                N=128, K=100000, tol=1e-12, drag=0.5, steep=200,
                output=False):
    # steady state
    k_ss = (alpha/(rho+delta))**(1/(1-alpha))
    s_ss = alpha*(delta/(rho+delta))

    # grid it up
    k_lo, k_hi = m_lo*k_ss, m_hi*k_ss
    k_grid = np.linspace(k_lo, k_hi, N)
    k_diff = k_grid[1:] - k_grid[:-1]

    # extras stuff
    y = k_grid**alpha

    # construct derivative matrices - be very very careful
    delvp = sp.diags(np.r_[-1/k_diff, 0]) + sp.diags(1/k_diff, offsets=1)
    delvm = sp.diags(np.r_[0, 1/k_diff]) + sp.diags(-1/k_diff, offsets=-1)
    delvd = delvp - delvm

    # initial guess
    inv = (s_ss*y)*np.ones(N) # k_dot = 0

    # iterate
    for j in range(K):
        # find derivative
        k_dot = inv - delta*k_grid
        q = sigmoid(steep*k_dot)
        q[0], q[-1] = 1, 0 # so we never address bounds

        # combine upper and lower
        delv = delvm + sp.diags(q)*delvd # need to be tricky

        # construct full value function equation
        A = rho*sp.eye(N) - sp.diags(k_dot)*delv
        b = util(y-inv, sig=sigma)

        # solve it!!!!
        v = sp.linalg.spsolve(A, b)
        dv = delv*v
        err = np.max(np.abs(A*v-b))

        # find investment
        c = dutil_inv(np.maximum(0, dv), sig=sigma)
        inv_new = y - c

        # convergence
        diff = np.max(np.abs(inv_new-inv))
        if diff < tol:
            if output:
                print(f'{j}: {err} {diff}')
                print('converged')
            break

        # output
        if output and j % 1000 == 0:
            print(f'{j}: {err} {diff}')

        # update
        inv = drag*inv + (1-drag)*inv_new

    # finish up
    u = util(c, sig=sigma)

    # return vectors
    return {
        'k': k_grid,
        'v': v,
        'i': inv,
        'dk': k_dot,
        'dv': dv,
        'c': c,
        'u': u
    }

def cont_cheb(rho=0.05, alpha=0.3, delta=0.1, sigma=2.0,
              N=32, K=100, tol=1e-12, drag=0.0, output=False):
    # steady state
    k_ss = (alpha/(rho+delta))**(1/(1-alpha))
    s_ss = alpha*(delta/(rho+delta))

    # change of variable
    k_lo, k_hi = m_lo*k_ss, m_hi*k_ss
    kofx = lambda x: k_lo + ((x+1)/2)*(k_hi-k_lo)
    xofk = lambda k: 2*((k-k_lo)/(k_hi-k_lo)) - 1
    dxdk = 2/(k_hi-k_lo)
    dkdx = (k_hi-k_lo)/2

    # base and mapped grid
    x_grid = np.cos(np.pi*((2*np.arange(N)+1)/(2*N)))[::-1] # chebyshev nodes
    k_grid = kofx(x_grid)
    y_grid = k_grid**alpha

    # matrix inputs - very very yep
    vander = cheb.chebvander(x_grid, N-1) # [N, D]
    dermap = np.vstack([cheb.chebder(np.eye(N), scl=dxdk), np.zeros((1, N))]) # [D, D]
    dvander = np.dot(vander, dermap) # [N, D]

    # initial guess
    inv = (s_ss*y_grid)*np.ones(N) # kdot = 0

    # loop until iteration
    for j in range(K):
        k_dot = inv - delta*k_grid

        A = rho*vander - k_dot[:, None]*dvander
        b = util(y_grid-inv, sigma)
        x = np.linalg.solve(A, b)

        v = np.dot(vander, x)
        dv = np.dot(dvander, x)

        c = dutil_inv(dv, sigma)
        inv_new = y_grid - c

        diff = np.max(np.abs(inv_new-inv))
        if output and j % 10:
            print(f'{j}: {diff}')

        if diff < tol:
            if output:
                print(f'{j}: {diff}')
                print('converged')
            break

        inv = drag*inv + (1-drag)*inv_new

    # finish up
    u = util(c, sig=sigma)

    # return vectors
    return {
        'k': k_grid,
        'v': v,
        'i': inv,
        'dk': k_dot,
        'dv': dv,
        'c': c,
        'u': u
    }

# check errors on a higher interpolated grid
def cont_errors(k0_grid, val, inv, dval, N1, kind='linear',
                rho=0.05, alpha=0.3, delta=0.1, sigma=2.0):
    # steady state
    k_ss = (alpha/(rho+delta))**(1/(1-alpha))
    k_lo, k_hi = m_lo*k_ss, m_hi*k_ss

    # grid it up
    k1_grid = np.linspace(k_lo, k_hi, N1)

    # interpolated values
    val_int = interp(k1_grid, k0_grid, val)
    inv_int = interp(k1_grid, k0_grid, inv)
    dv_int = interp(k1_grid, k0_grid, dval)

    # equation error
    y1 = k1_grid**alpha
    val1 = util(y1-inv_int, sig=sigma) + (inv_int-delta*k1_grid)*dv_int
    verr = rho*val_int - val1

    return k1_grid, verr

# simulate C/K path
def simulate_ck(k0, c0, k_lo, k_hi,
                rho=0.05, alpha=0.3, delta=0.1, sigma=2.0,
                dt=0.01, S=10000, direc=-1):
    # initial values
    k, c = k0, c0
    kpath, cpath = [k], [c]

    for t in range(S):
        # growth rates
        gk = k**(alpha-1) - delta - c/k
        gc = (1/sigma)*(alpha*k**(alpha-1)-(rho+delta))

        # adaptive step size
        dt1 = direc*dt
        k *= 1 + dt1*gk
        c *= 1 + dt1*gc

        # store
        kpath.append(k)
        cpath.append(c)

        # terminal condition
        if k <= k_lo:
            break
        if k >= k_hi:
            break

    return np.array(kpath), np.array(cpath)

def cont_reverse(rho=0.05, alpha=0.3, delta=0.1, sigma=2.0,
                 N=128, ptol=1e-4, output=False):
    # steady state
    k_ss = (alpha/(rho+delta))**(1.0/(1.0-alpha))
    i_ss = delta*k_ss
    y_ss = k_ss**alpha
    c_ss = y_ss - i_ss

    # grid it up
    k_lo, k_hi = m_lo*k_ss, m_hi*k_ss
    k_grid = np.linspace(k_lo, k_hi, N)

    # generate upper and lower paths
    kvec1, cvec1 = simulate_ck(k_ss-ptol, c_ss-ptol, k_lo, k_hi)
    kvec2, cvec2 = simulate_ck(k_ss+ptol, c_ss+ptol, k_lo, k_hi)

    # combined path
    kvec = np.concatenate([kvec1[::-1], kvec2])
    cvec = np.concatenate([cvec1[::-1], cvec2])

    # interpolate
    c = interp(k_grid, kvec, cvec)

    # other stuff
    y = k_grid**alpha
    i = y - c
    dk = i - delta*k_grid
    u = util(c, sig=sigma)

    # return vectors
    return {
        'k': k_grid,
        'i': i,
        'dk': dk,
        'c': c,
        'u': u
    }
